#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
     Main functions for obsinfo-validate. 
     
     obsinfo strongly follows the hierarchy of StationXML files.
       
     Module contains both class ValidateObsinfo and a main entry point for obsinfo-validate
"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA @UnusedWildImport

import os
from pathlib import Path, PurePath
import glob
import inspect
import difflib
import re
import sys
from json.decoder import JSONDecodeError
from argparse import ArgumentParser
from ..misc.configuration import ObsinfoConfiguration

import logging
from logging.handlers import RotatingFileHandler


#Third party imports
from obspy.core.utcdatetime import UTCDateTime

# obsinfo imports
from obsinfo.obsMetadata.obsmetadata import (ObsMetadata)
from obsinfo.instrumentation import (Instrumentation, InstrumentComponent,
                                     Datalogger, Preamplifier, Sensor,
                                     ResponseStages, Stage, Filter, Location)
from obsinfo.network import (Station, Network)
from obsinfo.instrumentation.filter import (Filter, PolesZeros, FIR, Coefficients, ResponseList,
                     Analog, Digital, AD_Conversion)
from obsinfo.misc.printobs import  (PrintObs)
from obsinfo.misc.discoveryfiles import Datapath
import obsinfo 
from obsinfo.obsMetadata.obsmetadata import ObsMetadata
from obsinfo.misc.discoveryfiles import Datapath
from obsinfo.misc.const import *


class ValidateObsinfo():
    """
    Contains methods to validate each level of information files.
    
    Attributes:
        datapath (:class:`/.Datapath`): Store datapath list(s)
        verbose (bool): Prints progress messages
        remote (bool): Indicates file in command line argument is to be found
            using datapath 
        debug (bool): Print more detailed messages. 
    """
    
    def setUp(self, verbose=True, remote=False, debug=False):
        """
        Sets up status variables according to .obsinforc and command line arguments.
        
        Args:
            verbose (bool): Print several progress messages
            remote (bool): Find file in command line argument using datapath
            debug (bool): Print more detailed messages and enable traceback
                of exceptions
        Returns: self
        """
        dp = Datapath()
        self.datapath = dp
        self.datapath.infofiles_path = dp.datapath_list 
        self.datapath.validate_path = Path(obsinfo.__file__).parent.joinpath('data', 'schemas')
    
        self.verbose = verbose
        self.remote = remote
        self.debug = debug
        
        return self

    def assertTextFilesEqual(self, first, second, msg=None):
        """
        Compares two text files
        
        :param first: First file to compare
        :type first: str
        :param second: Second file to compare
        :type second: str
        :param msg: Message to print in case of failure
        :type msg: str
        """
        with open(first) as f:
            str_a = f.read()
        with open(second) as f:
            str_b = f.read()
        if str_a != str_b:
            first_lines = str_a.splitlines(True)
            second_lines = str_b.splitlines(True)
            delta = difflib.unified_diff(
                first_lines, second_lines,
                fromfile=first, tofile=second)
            message = ''.join(delta)
            if msg:
                message += " : " + msg
            self.fail("Multi-line strings are unequal:\n" + message)
        
    def validate_all_filters(self):
        """
         Validate all information files in <component>/responses/_filters" subdirectory, as given by datapath.

         If you wish to test individual files, use test_filter(file) with file an absolute or 
         relative file name.
        """
        for dir in self.datapath.infofiles_path:
            files_in_validate_dir = Path(dir).joinpath(
                "*rs", # includes sensors, preamplifiers and dataloggers
                "responses",
                "filters/*.yaml")
            
            filelist = glob.glob(str(files_in_validate_dir))
            for file in filelist:
                self.validate_filter(file)
        
    def validate_filters_in_directory(self, dir):
        """
        Validate all information files in filter directory.
        
        :param dir: directory where filter files reside 
        :type dir: str
        """
        if dir.is_dir():
            for file in dir.iterdir():
                self.validate_filter(file)     

    def validate_filter(self, info_file):
        """
        Validate a single filter file.

        :param info_file: filter info file to validate. No assumptions made about path.
        :type info_file: str
        """
        ret = ObsMetadata().validate(self.datapath.validate_path, info_file,
                                     self.remote, "yaml", "filter",
                                     self.verbose, "filter", False)            
        if self.verbose: 
            self._print_passed(f'Filter test for: {info_file}', ret)      
        
    def validate_all_stages(self):
        """
        Validate all information files in each responses subdirectory "<component>/responses/_filters"
        
        If you wish to test individual files, use test_stage(file) with file an absolute or 
        relative file name.
        """
        for dir in self.datapath.infofiles_path:
            files_in_validate_dir = Path(dir).joinpath(
                "*rs", # includes sensors, preamplifiers and dataloggers
                "responses/*.yaml"
                )
            filelist = glob.glob(str(files_in_validate_dir))
            for file in filelist:
                self.validate_stage(file)
        
    def validate_stages_in_directory(self, dir):
        """
        Validate all information files in stage directory as given by datapath.
        
        :param dir: directory where stage files reside 
        :type dir: str
        """              
        datalogger_dir_re = re.compile(".*/dataloggers")
        sensor_dir_re = re.compile(".*/sensors")
        exception_re = re.compile(".*/test-with")
        if re.match(datalogger_dir_re, str(dir)) or re.match(sensor_dir_re, str(dir)):
            for file in (Path(dir).joinpath("responses")).iterdir():
                if not file.is_dir() and re.match(exception_re, str(file)) == None: 
                    self.validate_stage(file)        
        
    def validate_stage(self, info_file):
        """
        Validate a single stage file.

        :param info_file: stage info file to validate. No assumptions made about path.
        :type info_file: str
        """   
        ret = ObsMetadata().validate(self.datapath.validate_path,
                                     info_file, self.remote, "yaml", "stage",
                                     self.verbose, "stage", False)         
        if self.verbose: 
            self._print_passed(f'Stage validate for: {info_file}', ret)      
            
        
    def validate_all_components(self):
        """
        Validate all information files in each components (sensor, preamplifier, datalogger) subdirectory
        as given by datapath
        """
        components_list = ["sensors",
                           "preamplifiers",
                           "dataloggers"]
        for dir in self.datapath.infofiles_path:
            for comp in components_list:
                # includes sensors, preamplifiers and dataloggers
                files_in_validate_dir = Path(dir).joinpath(comp, "*.yaml")
                filelist = glob.glob(str(files_in_validate_dir))
                for file in filelist:
                    ret = ObsMetadata().validate(self.datapath.validate_path,
                                                 file, self.remote, "yaml",
                                                 comp, self.verbose, comp[:-1],
                                                 False)
                    if self.verbose: 
                        self._print_passed(file, ret)  

    def validate_datalogger(self, info_file):
        """
        Validate a single datalogger file.

        :param info_file: datalogger info file to validate. No assumptions made about path.
        :type info_file: str
        """
        # OJO: no configuraton passed from above. No delay_correction either.
        ret = ObsMetadata().validate(self.datapath.validate_path, info_file,
                                     self.remote, "yaml", "datalogger",
                                     self.verbose, "datalogger", False)
        if self.verbose: 
            self._print_passed(info_file, ret)  
           
    def validate_sensor(self, info_file):
        """
        Validate single sensor instrument_component information file
        
        :param info_file: sensor info file to validate. No assumptions made about path.
        :type info_file: str
        """
        #OJO: no configuraton passed from above. No delay_correction either.
        ret = ObsMetadata().validate(self.datapath.validate_path, info_file,
                                     self.remote, "yaml", "sensor",
                                     self.verbose, "sensor", False)
        if self.verbose: 
            self._print_passed(info_file, ret)  
        
    def validate_preamplifier(self, info_file):
        """
        Validate single preamplifier instrument_component information file
        
        :param info_file: preamplifier info file to validate. No assumptions made about path.
        :type info_file: str
        """
        #OJO: no configuraton passed from above. No delay_correction either.
        ret = ObsMetadata().validate(self.datapath.validate_path, info_file,
                                     self.remote, "yaml", "preamplifier",
                                     self.verbose, "preamplifier", False)
        if self.verbose: 
            self._print_passed(info_file, ret)  
           
    def validate_all_instrumentations(self):
        """
        Validate all information files for instrumentations in a directory given by datapath
        
        If you want to validate a single instrumentation file use validate_instrumentation.
        """
        
        for dir in self.datapath.infofiles_path:
            # includes sensors, preamplifiers and dataloggers
            files_in_validate_dir = Path(dir).joinpath("instrumentation/*.yaml")
            filelist = glob.glob(str(files_in_validate_dir))
            for file in filelist:
                self.validate_instrumentation(file)      

    def validate_instrumentation(self, info_file):
        """
        Validate single instrumentation information file
        
        :param info_file: instrumentation info file to validate. No assumptions
            made about path.
        :type info_file: str
        """
        ret = ObsMetadata().validate(self.datapath.validate_path, info_file,
                                     self.remote, "yaml", "instrumentation",
                                     self.verbose,
                                     "instrumentation.schema.json", False)
        if self.verbose: 
            self._print_passed(info_file, ret)  
       
    @staticmethod 
    def _print_passed(text, passed):
        if passed:   
            print(f'{text}: PASSED')
        else:   
            print(f'{text}: FAILED')
                     
    def validate_station(self, info_file='TEST.station.yaml', level="all"):
        """
        Validate single station information file
        
        :param info_file: station info file to validate. No assumptions made about path.
        :type info_file: str
        """
        ret = ObsMetadata().validate(self.datapath.validate_path, info_file,
                                     self.remote, "yaml", "station",
                                     self.verbose, "station.yaml.json", False)   
        if self.verbose: 
            self._print_passed(f'Station validate for: {info_file}', ret)  

    def validate_all_networks(self):
        """
        Validate all network information files in a directory.given by datapath
        """
        
        for dir in self.datapath.infofiles_path:
            # includes sensors, preamplifiers and dataloggers
            files_in_validate_dir = Path(dir).joinpath("network/*.yaml")
            filelist = glob.glob(str(files_in_validate_dir))
            for file in filelist:
                self.validate_network(file)

    def validate_network(self, info_file):
        """
        Validate single network information file
        
        :param info_file: network info file to validate. No assumptions made about path.
        :type info_file: str
        """
        ret = ObsMetadata().validate(self.datapath.validate_path, info_file, "yaml", "network", self.remote, self.verbose, "network", False)
        if self.verbose: 
            self._print_passed(f'Network validate for: {info_file}', ret)  

  
def main():
    """
    Entry point for obsinfo-validate. 
     
     1) Setups status variables from command line arguments.
     2) Validates file according to file type contained in name
     3) Manages all uncaught exceptions
    """
    args = retrieve_arguments()
    input_filename = args.input_filename
    logger = init_logging()
    
    val = ValidateObsinfo()
    val.setUp(verbose=args.verbose, remote=args.remote, debug=args.debug)
    
    if args.schema:
        if is_valid_type(args.schema):
            type = args.schema
        else:
            raise ValueError(f'Unknown schema type: {args.schema}')
    else:
        type = ObsMetadata.get_information_file_type(input_filename)
         
    if args.verbose:
        print(f'Validating {type} file')
            
#     try:               
    if type == "filter":
        val.validate_filter(input_filename)
    elif type == "stage":
        val.validate_stage(input_filename)
    elif type == "datalogger":
        val.validate_datalogger(input_filename)
    elif type == "preamplifier":
        val.validate_preamplifier(input_filename)
    elif type == "sensor":
        val.validate_sensor(input_filename)
    elif type == "instrumentation":
        val.validate_instrumentation(input_filename)
    elif type == "station":
        val.validate_station(input_filename)
    elif type == "network":
        val.validate_network(input_filename)
                       
#     except TypeError as e:
#         print("TypeError:" + str(e))
#         logger.error("TypeError: Illegal format: fields may be missing or with wrong format in input file, or there is a programming error")
#         if args.debug:
#             raise
#         sys.exit(EXIT_DATAERR)
#     except (KeyError, IndexError) as e:
#         print("Illegal value in dictionary key or list index:" + str(e))
#         logger.error("KeyError, IndexError: Illegal value in dictionary key or list index")
#         if args.debug:
#             raise
#         sys.exit(EXIT_SOFTWARE)
#     except ValueError as e:
#         print("ValueError:" + str(e))
#         logger.error("ValueError: An illegal value was detected")
#         if args.debug:
#             raise
#         sys.exit(EXIT_DATAERR)
#     except FileNotFoundError as e:
#         if args.debug:
#             raise
#         print("File not found:" + str(e))
#         logger.error(f"FileNotFoundError: {str(e)}")
#         sys.exit(EXIT_NOINPUT)
#     except JSONDecodeError as e:
#         print("JSONDecodeError:" + str(e))
#         logger.error("JSONDecodeError: File and/or subfiles have an illegal format. Probably indentation or missing quotes/parentheses/brackets")
#         if args.debug:
#             raise 
#         sys.exit(EXIT_DATAERR)
#     except (IOError, OSError, LookupError) as e:
#         print("File could not be opened or read:" + str(e))
#         logger.error("IOError, OSError, LookupError: File could not be opened or read")
#         if args.debug:
#             raise
#         sys.exit(EXIT_UNAVAILABLE)
#     except AttributeError as e:
#         print("Attribute error: " + str(e))
#         logger.debug("AttributeError: Programming error: an object in code had a wrong attribute")
#         if args.debug:
#             raise
#         sys.exit(EXIT_SOFTWARE)
#     except:
#         print("General exception")
#         logger.debug("General exception:")
#         if args.debug:
#             raise
#         sys.exit(EXIT_FAILURE)
#    
    sys.exit(EXIT_SUCCESS)
     

def retrieve_arguments():
    """
    Retrieve arguments from command line. Setup several status variables and get information file name
      
    :returns: dictionary object with all status variables and information file name.
    """
    # Parse the arguments
    parser_args = ArgumentParser(prog="obsinfo-validate")
    
    #flags
    parser_args.add_argument(
        "-q", "--quiet", action='store_true', default=False,
        help="Quiet operation. Don't print informative messages") 
    parser_args.add_argument(
        "-s", "--schema", help="Force validation against given schema") 
    parser_args.add_argument(
        "-r", "--remote",  action='store_true', default=False,
        help="Search input_filename in the DATAPATH repositories")
    parser_args.add_argument(
        "-d", "--debug",  action='store_true', default=False,
        help="Print traceback for exceptions")  
    # positional arguments
    parser_args.add_argument("input_filename", type=str,
                             help="Information file to be validated.")
    
    args = parser_args.parse_args()
    args.verbose = not args.quiet
    
    if not Path(args.input_filename).is_absolute():
        args.input_filename = str(Path(os.getcwd()).joinpath(args.input_filename).resolve())
             
    return args  


def init_logging():
    """
    Create or open a rotating logging file and add it to ObsinfoConfiguration
    
    :returns: object of Logger class 
    """
    logfile = Path.home().joinpath('.obsinfo', 'obsinfolog-validate')
    
    logger = logging.getLogger("obsinfo")
    
    logger.setLevel(logging.DEBUG)
    # add a rotating handler with 200K (approx) files and just two files
    handler = RotatingFileHandler(logfile, maxBytes=200000,
                                  backupCount=2)
    frmt = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    handler.setFormatter(frmt)
    logger.addHandler(handler)  
    
    return logger 

    
if __name__ == '__main__':
    main()



"""
InstrumentComponent class and subclasses Sensor, Preamplifier, Datalogger.
Equipment class
"""
# Standard library modules
import warnings
import logging

# Non-standard modules
from obspy.core.inventory.util import Equipment as obspy_Equipment
from obspy.core.utcdatetime import UTCDateTime

# obsinfo
from obsinfo.obsMetadata.obsmetadata import (ObsMetadata)

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Equipment(obspy_Equipment):
    """
    Equipment.

    Equivalent to :class: obspy.core.inventory.util.Equipment

    Attributes:
        type (str):
        channel_modif (str):
        selected_config (str):
        description (str):
        manufacturer (str):
        model (str):
        vendor (str):
        serial_number (str):
        installation_date (str in date format):
        removal_date (str in date format):
        calibration_dates (str in date format):
        resource_id (str):
        obspy_equipment (class `obspy.core.inventory.equipmentEquipment`)`
    """

    def __init__(self, attributes_dict, channel_modif={}, selected_config={}):
        """
        Constructor

        Args:
            attributes_dict (dict or :class:`ObsMetadata`): attributes of
                component, hopefully including configuration_default
            channel_modif (dict or :class:`ObsMetadata`): channel modifications
                inherited from station
            selected_config (sdict or :class:`ObsMetadata`): configuration dict
                selected at instrument_component level

        """
        self.type = attributes_dict.get_configured_element(
            'type', channel_modif, selected_config, None)
        self.description = attributes_dict.get_configured_element(
            'description', channel_modif, selected_config, None)
        self.manufacturer = attributes_dict.get_configured_element(
            'manufacturer', channel_modif, selected_config, None)
        self.model = attributes_dict.get_configured_element(
            'model', channel_modif, selected_config, None)
        self.vendor = attributes_dict.get_configured_element(
            'vendor', channel_modif, selected_config, None)
        self.serial_number = attributes_dict.get_configured_element(
            'serial_number', channel_modif, selected_config, None)
        self.installation_date = ObsMetadata.validate_date(
            attributes_dict.get_configured_element('installation_date',
                                                   channel_modif,
                                                   selected_config,
                                                   None))
        self.removal_date = ObsMetadata.validate_date(
            attributes_dict.get_configured_element('removal_date',
                                                   channel_modif,
                                                   selected_config,
                                                   None))
        calibration_dates = ObsMetadata.validate_dates(
            attributes_dict.get_configured_element('calibration_dates',
                                                   channel_modif,
                                                   selected_config,
                                                   []))

        self.calibration_dates = calibration_dates
        self.resource_id = None

        equip = self.to_obspy()
        self.obspy_equipment = equip[0] if isinstance(equip, tuple) else equip

    def __repr__(self):
        s = 'Equipment('
        if self.type:
            s += f', Type={self.type}'
        if self.description:
            s += f', Description={self.description}'
        if self.manufacturer:
            s += f', Manufacturer={self.manufacturer}'
        if self.model:
            s += f', Model={self.model}'
        if self.vendor:
            s += f', Vendor={self.vendor}'
        if self.serial_number:
            s += f', Serial Number={self.serial_number}'
        if self.installation_date:
            s += f', Installation Date={self.installation_date}'
        if self.removal_date:
            s += f', Removal Date={self.removal_date}'
        if self.calibration_dates:
            s += f', Calibration Date={self.calibration_dates}'
        s += ')'
        return s

    def to_obspy(self):
        """
        Convert an equipment (including the equipment description in
        components) to its obspy object

        Returns:
            (:class:`obspy.core.invertory.util.Equipment`)
        """
        resource_id = None

        installation_date = UTCDateTime(self.installation_date) \
            if self.installation_date else None
        removal_date = UTCDateTime(self.removal_date) \
            if self.removal_date else None

        if isinstance(self.calibration_dates, list) \
                and len(self.calibration_dates) > 0:
            calib_dates = [UTCDateTime(dt)
                           for dt in self.calibration_dates]
        else:
            calib_dates = []

        equip = obspy_Equipment(self.type, self.description, self.manufacturer,
                                self.vendor, self.model, self.serial_number,
                                installation_date, removal_date,
                                calib_dates, resource_id)

        return equip



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Test network and station classes
"""
import warnings
from pathlib import Path
import unittest
import difflib

# Third party  imports

# obsinfo modules
from obsinfo.network import (Station, Network)

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
verbose = False


class NetworkTest(unittest.TestCase):
    """
    Class of test methods for network and station objects

    Attributes:
        testing_path (str): path to datafiles to be tested aside from the
            examples
        level (str): level to be printed
        test (boolean): determines if this is test mode
        print_output (boolean): determines if this is print mode. Both can
            coexist.
    """

    def setUp(self, test=True, print_output=False, level=None):
        """
        Set up default values and paths
        """
        self.testing_path = Path(__file__).parent.joinpath("data")

        self.level = level
        self.test = test
        self.print_output = print_output

    def assertTextFilesEqual(self, first, second, msg=None):
        with open(first) as f:
            str_a = f.read()
        with open(second) as f:
            str_b = f.read()

        if str_a != str_b:
            first_lines = str_a.splitlines(True)
            second_lines = str_b.splitlines(True)
            delta = difflib.unified_diff(
                first_lines, second_lines,
                fromfile=first, tofile=second)
            message = ''.join(delta)
            if msg:
                message += " : " + msg
            self.fail("Multi-line strings are unequal:\n" + message)

    def test_read_network_dict(self):
        """Tests reading a network dictionary"""
        station={"site": "bob",
                 "location_code":"00",
                 "locations": 
                    {"00": 
                        {"base": 
                            {"uncertainties.m": {"lat":0, "lon": 0, "elev": 0},
                             "depth.m": 0,
                             "geology":"rock",
                             "vault":"seafloor"
                            },
                        "position":
                            {"lat": 0, "lon": 0, "elev":0
                            }
                        }
                    }
                }
        net_dict = {"network_info": {"code": '4G',
                                     "name": 'test network',
                                     "start_date": "2021-01-01",
                                     "end_date": "2021-01-02",
                                     "description": "Oh what a night!",
                                     },
                    "campaign_ref_name": "HOHO!",
                    "operator": {"full_name": "My park", "reference_name": "INSU-IPGP"},
                    "stations": {"howdy": station.copy(),
                                 "doody": station.copy()}
                   }
        net = Network(net_dict, station_only=True)

def suite():
    return unittest.makeSuite(NetworkTest, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')




"""
Application to configure variables used by obsinfo in file .obsinforc
and to copy examples and templates out of the distribution directories

"""

import os
import sys
import re
import site
import shutil
import datetime
import platform
from pathlib import Path
import logging


from argparse import ArgumentParser

import obsinfo
from ..misc.const import *
from obsinfo.version import __version__

def setup_obsinfo():
    """
    Entry point to configure variables in file .obsinforc
    
    and to copy examples and templates out of the distribution directories
    """
    
    
    # initialize variables
    version = platform.python_version()    
    path_to_package = Path(obsinfo.__file__).parent
    examples_dir = path_to_package.joinpath("_examples/Information_Files")
    templates_dir = path_to_package.joinpath("templates")     
    home = Path.home()
    configuration_file = home.joinpath('.obsinforc')
    
    args = retrieve_arguments()
    
    validate_arguments(args)  
    
    # create logging directory if it doesn't exist
    logdir = Path().home().joinpath('.obsinfo').resolve()
    if not os.path.isdir(logdir):
        os.mkdir(logdir)  
    
    # copy examples and templates
    if args.dest and not args.no_copy:
        try:
            dest = Path(args.dest).resolve()
            # if dest dir exists rename it with timestamp in order to preserve it
            if os.path.isdir(dest) or os.path.isfile(dest):
                new_dest = str(dest) + "." + str(datetime.datetime.today())
                os.rename(dest, new_dest)
            
            shutil.copytree(examples_dir, dest.joinpath("Information_Files"))
            shutil.copytree(templates_dir, dest.joinpath("templates"))
        except NotADirectoryError:
            print("Directory not found")
            exit(EXIT_IOERR)
        except OSError:
            print("Operating system error")
            exit(EXIT_IOERR)
    # create configuration file
    output=[]
    
    output.append("gitlab_repository: " +  args.gitlab + "\n")
    output.append("gitlab_project: " + args.project + "\n")
    output.append("gitlab_path: " + args.remote_path + "\n")
    pat = re.compile("^[0-9]+\.[0-9]+")
    version = pat.match(__version__)
    branch = "v" + version.group(0) if args.branch else "master"
        
    output.append('obsinfo_branch: ' + branch + '\n')
    
    remote_repository = args.gitlab + "/" + args.project + "/" + args.remote_path \
          if not args.no_remote else ""
    
    s = build_datapath_list(args, remote_repository)
    if s:
        output.append(s + "\n")
        
    try:
        with open(configuration_file, "w") as fp:
           
          fp.writelines(output)
    
    except OSError:
        warnings.warn("Could not create configuration file. Fix problem and run obsinfo-setup again\n")
        exit(EXIT_IOERR) 
        
        
def retrieve_arguments():
    """
    Parse command line arguments
    
    :returns: dictionary of arguments, both status and directory names 
    
    """
    
    # Parse the arguments
    parser_args = ArgumentParser(prog="obsinfo-setup")
    
    #flags
    parser_args.add_argument("-x", "--no_examples", action='store_true', default=False,
                             help="Don't import examples, only templates, and remove examples directory from the datapath"
                            ) 
    
    parser_args.add_argument("-c", "--no_copy",  action='store_true', default=False,
                             help="Don't import anything at all, don't create dest directory, which will be removed from datapath"
                            )  
    parser_args.add_argument("-n", "--no_remote",  action='store_true', default=False,
                             help="Install obsinfo without access to a gitlab repository.\n\
                             May be needed in some operating systems for compatibility issues",
                            )  
    parser_args.add_argument("-v", "--invert_datapath",  action='store_true', default=False,
                             help="Put remote gitlab repositorey first. All local directories will keep their order"
                            )
    parser_args.add_argument("-b", "--branch",  action='store_true', default=False,
                             help="Specifies the git branch to use, if not master"
                            )
   
    # optional arguments
    parser_args.add_argument("-d", "--dest", default=None,
                             help="Destination directory for templates and examples.",
                            )
    parser_args.add_argument("-g", "--gitlab", default="https://www.gitlab.com",
                             help="Gitlab repository)",
                            )
    parser_args.add_argument("-p", "--project", default="resif/obsinfo",
                             help="path to project and the directory where information files lie within the Gitlab repository)",
                            )
    parser_args.add_argument("-l", "--local_repository", default=None,
                             help="Specify local repository for information files and include it as first or second option in datapath"
                            )
    parser_args.add_argument("-w", "--working_directory", default=None,
                             help="Specify working directory for obsinfo and include it as first option in datapath")
    
    parser_args.add_argument("-P", "--remote_path", default="obsinfo/_examples/Information_Files",
                             help="Specify remote directory under project"
                            ) 
           
    return parser_args.parse_args()   

def validate_arguments(args):
    """
    Validate that directories in command line arguments exist. Raise an exception if not.
    
    :param args: directory with command line argument, but status and directories
    :type args: dictionary of strings
    
    :raises OSError, ValueError
    
    """
    
    if args.dest:
        if not Path(args.dest):
            raise OSError(ENOENT, "Path for example and template destination directory is an illegal path")
    else:
        default_information_files = ""
    
    if args.local_repository:
        if not Path(args.local_repository):
            raise OSError(ENOENT, "Path for local repository for information files is an illegal path")
    
    if args.working_directory:
        if not Path(args.working_directory):
            raise OSError(ENOENT, "Path for obsinfo working directory is an illegal path")
    
    if not args.no_remote:
        if not Path(args.gitlab).as_uri:
            raise ValueError('Gitlab repository must be specified as a legal URI unless "-n" is on')
    
    if not args.no_remote:
        if not Path(args.project):
            raise ValueError('Project must be specified as a legal URI unless "-n" is on')
                    
def build_datapath_list(args, remote_repository):
    """
    Build datapath as a list of directories to search, similar to the PATH environment variable.
    
    #. Add working_directory as first option if it exists
    #. Next add local repository first if it exists (wll be first if no working directory specified)
    #. Finally add destination, where standard examples lie. This won't be used if -x is on
    #. Final directory will be the remote one unless -v (invert) is specified
    
    """
    
    default_information_files = str(Path(args.dest).joinpath("Information_Files").resolve()) if args.dest else ""
    remote_first = remote_repository + ", " if args.invert_datapath else ""
    remote_last = remote_repository if not args.invert_datapath else ""
    
    s =  "datapath: ["
    s += remote_first 
    s += args.working_directory + ", " if args.working_directory else "" 
    s += args.local_repository + ", " if args.local_repository else ""
    s += default_information_files  + ", " if default_information_files and not args.no_examples else ""
    s += remote_last if remote_last else ""
    s += "]"
    
    if not remote_last:
         s.replace(", ]", "]")

    return s            
        


"""
Location classe
"""
# Standard library modules
import warnings
import numpy as np
import logging

# Non-standard modules
from obspy.core.util.obspy_types import FloatWithUncertaintiesFixedUnit
from obspy.core.inventory.util import (Latitude, Longitude)
from obspy.geodetics.base import kilometers2degrees


warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Location(object):
    """
    Location class.

    Attributes:
        latitude (float): station latitude (degrees N)
        longitude (float): station longitude (degrees E)
        elevation (float): station elevation (meters above sea level)
        uncertainties_m (dict): 'lat', 'lon', 'elev' in METERS
        geology (str): site geology
        vault (str): vault type
        depth_m (float): depth of station beneath surface (meters)
        localisation_method (str): method used to determine position
        obspy_latitude: latitude as an *obspy* object
        obspy_longitude: longitude as an *obspy* object

    """

    def __init__(self, attributes_dict):
        """
        Create Location object and assign attributes from attributes_dict.
        Validate required location attributes exist
        Convert to obspy longitude and latitude

        :param attributes_dict: location information
        :type attributes_dict: dict or object of :class:`ObsMetadata`
        """
        position = attributes_dict.get('position', None)
        base = attributes_dict.get('base', None)

        if not base:
            msg = 'No base in location'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)
        if not position:
            msg = 'No position in location'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)

        self.latitude = position.get('lat', None)
        self.longitude = position.get('lon', None)
        self.elevation = position.get('elev', None)
        self.uncertainties_m = base.get('uncertainties.m', None)
        self.geology = base.get('geology', None)
        self.vault = base.get('vault', None)
        self.depth_m = base.get('depth.m', None)
        self.localisation_method = base.get('localisation_method', None)

    def __repr__(self):
        s = f'Location({self.latitude:g}, {self.longitude:g}, '
        s += f'{self.elevation:g}, {self.uncertainties_m}'
        if not self.geology == 'unknown':
            s += f', "geology: {self.geology}"'
        if self.vault:
            s += f', vault="{self.vault}"'
        if self.depth_m is not None:
            s += f', depth_m={self.depth_m:g}'
        if self.localisation_method:
            s += f', localisation_method="{self.localisation_method}"'
        s += ')'
        return s

    @property
    def obspy_elevation(self):
        """
        Returns elevation as obspy FloatWithUncertaintiesFixedUnit object
        """
        return FloatWithUncertaintiesFixedUnit(
            value=self.elevation,
            lower_uncertainty=self.uncertainties_m.get('elev', None),
            upper_uncertainty=self.uncertainties_m.get('elev', None),
            measurement_method=self.localisation_method)

    @property
    def obspy_latitude(self):
        """
        Returns obspy Latitude object
        """
        uncert_m = self.uncertainties_m.get('lat', None) \
            if self.uncertainties_m else None
        if uncert_m is None:
            lat_uncert = None
        else:
            meters_per_degree_lat = 1852.0 * 60.0
            # lat_uncert = kilometers2degrees(uncert_m / 1000)
            lat_uncert = uncert_m / meters_per_degree_lat
            lat_uncert = float("{:.3g}".format(lat_uncert))
        return Latitude(value=self.latitude,
                        lower_uncertainty=lat_uncert,
                        upper_uncertainty=lat_uncert)

    @property
    def obspy_longitude(self):
        """
        Returns obspy Longitude object
        """

        uncert_m = self.uncertainties_m.get('lon', None) \
            if self.uncertainties_m else None
        if uncert_m is None or abs(self.latitude) == 90:
            lon_uncert = None
        else:
            # lon is along parallel, not a great circle, so we must reduce by
            # cosine.
            meters_per_degree_lon = 1852.0 * 60.0 * np.cos(self.latitude * np.pi / 180.0)
            # lon_uncert = kilometers2degrees(
            #     (uncert_m / 1000) * np.cos(np.radians(self.latitude)))
            lon_uncert = uncert_m / meters_per_degree_lon
            lon_uncert = float("{:.3g}".format(lon_uncert))
        return Longitude(value=self.longitude,
                         lower_uncertainty=lon_uncert,
                         upper_uncertainty=lon_uncert)

    @staticmethod
    def get_location_from_code(locations, code, type, label):
        """
        Obtain from the locations dictionary one location by code (key).
        Raise exception if not found
        """
        if not locations:
            msg = 'No locations specified in station'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)

        loc = locations.get(code, None)

        if not loc:
            if type == 'station':  # station location code is mandatory
                msg = f'Location "{code}" not found in station {label}'
                warnings.warn(msg)
                logger.error(msg)
                raise TypeError(msg)
            else:
                msg = f'Channel {label} has no location code. Assuming '\
                      'location code "00"'
                warnings.warn(msg)
                logger.warning(msg)

                loc = locations.get("00", None)
                if not loc:
                    msg = f'Location "{code}" not found in channel {label}'
                    warnings.warn(msg)
                    logger.error(msg)
                    raise TypeError(msg)
        return loc



"""
Orientation class
"""
# Standard library modules
import warnings
import logging

# Non-standard modules
import obspy.core.util.obspy_types as obspy_types

# obsinfo modules
from obsinfo.obsMetadata.obsmetadata import (ObsMetadata)

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Orientation(object):
    """
    Class for sensor orientations. No channel modifs. Cannot change orientation
    as it is part of the channel identifiers. Azimuth and dip can be changed
    Orientation is coded by `FDSN standard <http://docs.fdsn.org/projects/
    source-identifiers/en/v1.0/channel-codes.html>`

    Attributes:
        azimuth (degrees): azimuth, clockwise from north
        azimuth_uncertainty (degrees) - For OBS, uncertainty is usually 180º
        dip (degrees): dip,  -90 to 90: positive=down, negative=up
        type dip_uncertainty (degrees)- For OBS, uncertainty is usually 180º
    """

    def __init__(self, attributes_dict, channel_modifs, polarity):
        """
        Constructor

        :Seismometer: If a positive voltage corresponds to upward motion
            (typical vertical seismometer), dip = -90º (up).  If positive
            voltage corresponds to downward motion (typical vertical geophone),
            dip = 90º
        :Hydrophone:  If a positive voltage corresponds to a positive
            (compressional) pressure, then dip = -90º.  If a positive voltage
            corresponds to a decrease in pressure, then dip = 90°

        Args:
            attributes_dict (dict or :class:`.ObsMetadata`): operator information
        """

        if not attributes_dict:
            msg = 'No orientation'
            warnings.warn(msg)
            logger.error(msg)
            raise ValueError(msg)

        # if a dictionary attributes_dict contains azimuth and/of dip info
        # else it's a simple string and is included in a list for generality
        keys = list(attributes_dict.keys()
                    if isinstance(attributes_dict, dict) else attributes_dict)
        if "1" in keys:
            self.orientation_code = "1"
            value = ObsMetadata(attributes_dict.get("1", None))
            if not value:
                msg = 'Type "1" channel has no azimuth'
                warnings.warn(msg)
                logger.error(msg)
                raise ValueError(msg)
            azimuth, azimuth_uncertainty = self.get_value_with_uncertainty(
                value.get_configured_element('azimuth.deg', channel_modifs,
                                             {}, None))
            if azimuth is None:
                msg = 'Type "1" channel has no azimuth'
                warnings.warn(msg)
                logger.error(msg)
                raise ValueError(msg)
            dip, dip_uncertainty = [0, None]
        elif "2" in keys:
            self.orientation_code = "2"
            value = ObsMetadata(attributes_dict.get("2", None))
            if not value:
                msg = 'Type "2" channel has no azimuth'
                warnings.warn(msg)
                logger.error(msg)
                raise ValueError(msg)
            azimuth, azimuth_uncertainty = self.get_value_with_uncertainty(
                value.get_configured_element('azimuth.deg',  channel_modifs,
                                             {}, None))
            if azimuth is None:
                msg = 'Type "2" channel has no azimuth'
                warnings.warn(msg)
                logger.error(msg)
                raise ValueError(msg)
            dip, dip_uncertainty = [0, None]
        elif "3" in keys:
            self.orientation_code = "3"
            value = ObsMetadata(attributes_dict.get("3", None))
            if not value:
                msg = 'Type "3" channel has no dip'
                warnings.warn(msg)
                logger.error(msg)
                raise ValueError(msg)
            azimuth, azimuth_uncertainty = [0, None]
            dip, dip_uncertainty = self.get_value_with_uncertainty(
                value.get_configured_element('dip.deg', channel_modifs,
                                             {}, None))
            if dip is None:
                msg = 'Type "3" channel has no dip'
                warnings.warn(msg)
                logger.error(msg)
                raise ValueError()
        elif "H" in keys:
            self.orientation_code = "H"
            azimuth, azimuth_uncertainty = [0, None]
            dip, dip_uncertainty = [90 * polarity, 5]
        elif "X" in keys:
            self.orientation_code = "X"
            azimuth, azimuth_uncertainty = [0, 5]  # Defined by FDSN
            dip, dip_uncertainty = [0, None]
        elif "Y" in keys:
            self.orientation_code = "Y"
            azimuth, azimuth_uncertainty = [90, 5]  # Defined by FDSN
            dip, dip_uncertainty = [0, None]
        elif "Z" in keys:
            self.orientation_code = "Z"
            azimuth, azimuth_uncertainty = [0, None]  # Defined by FDSN
            dip, dip_uncertainty = [90 * polarity, 5]
        else:
            msg = 'Type(s) "{keys}" orientation subcode(s) not implemented'
            warnings.warn(msg)
            logger.error(msg)
            raise ValueError()

        self.azimuth = obspy_types.FloatWithUncertaintiesAndUnit(
            azimuth, lower_uncertainty=azimuth_uncertainty,
            upper_uncertainty=azimuth_uncertainty, unit='degrees')
        self.dip = obspy_types.FloatWithUncertaintiesAndUnit(
            dip, lower_uncertainty=dip_uncertainty,
            upper_uncertainty=dip_uncertainty, unit='degrees')

    def __repr__(self):
        s = '\nOrientation('
        if self.azimuth:
            s += f', Azimuth={self.azimuth}'
        if self.dip:
            s += f', Dip={self.dip}'

        return s

    def get_value_with_uncertainty(self, info_list):
        """
        Validate that info_list is a 2 member list.

        Args:
            info_list (list of strings or floats): two numbers, a value and
                an uncertainty
        Returns:
            (list): the two members
        Raises:
            ValueError if not a list of two floats or strings that can be
                converted to floats
        """

        try:
            if not info_list or not isinstance(info_list, list) \
                    or len(info_list) != 2 \
                    or not isinstance(float(info_list[0]), float):
                msg = f"value and uncertainty or both are illegal {info_list}"
                warnings.warn(msg)
                logger.error(msg)
                raise ValueError(msg)
        except ValueError:
            msg = f"value or uncertainty or both are illegal {info_list}"
            warnings.warn(msg)
            logger.error(msg)
            raise ValueError(msg)
        if info_list[1] is not None:
            if not isinstance(float(info_list[1]), float):
                msg = f"value and uncertainty or both are illegal {info_list}"
                warnings.warn(msg)
                logger.error(msg)
                raise ValueError(msg)
        return [info_list[0], info_list[1]]



"""
Instrumentation class and all subclasses

nomenclature:
    * An "Instrument" (measurement instrument) records one physical parameter
    * A "Channel" is an Instrument + an orientation code and possibly
        starttime, endtime and location code
    * An "Instrumentation" combines one or more measurement Channels
"""
from .instrumentation import Instrumentation
from .instrument_component import (InstrumentComponent, Datalogger, Sensor,
                                   Preamplifier)
from .equipment import Equipment
from .channel import Channel
from .filter import Filter
from .instrument import Instrument
from .location import Location
from .orientation import Orientation
from .response_stages import (ResponseStages, Stage)